
------------------------------------------------------------------------
--
-- AbsorbsMonitor
--
-- Copyright (C) 2010  Philipp Schmidt
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License
-- as published by the Free Software Foundation; either version 2
-- of the License, or (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to:
--
-- Free Software Foundation, Inc.,
-- 51 Franklin Street, Fifth Floor,
-- Boston, MA  02110-1301, USA.
--
--
------------------------------------------------------------------------


AbsorbsMonitor = LibStub("AceAddon-3.0"):NewAddon("AbsorbsMonitor", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0", "AceSerializer-3.0");


-- #################################################################################################


local playerGUID;
local playerClass;

-- Specifies the channel any AddOn message should be
-- sent to, nil if silent
-- Can also be used to get the last known party state
local curChatChannel = nil;

-- Always hold the timestamp from the last COMBAT_LOG_EVENT_UNFILTERED fired
local lastCombatLogEvent = 0.0;

-- Table of all active absorb effects indexed by GUID and then spellId
-- at spellId == -1 there is a numeric entry with the total remaining value,
-- at spellId == -2 there is a numeric entry with the total quality (that is, the minimal quality)
-- [GUID] = { [spellId] = {spellId, priority, remainingValue, maxValue, quality, durationTimerHandle, extra} }
local activeEffectsBySpell = {};

-- Table of all active absorb effects indexed by GUID and then in the order in which
-- they will be used
local activeEffectsByPriority = {};

-- Table of all active area absorb effects indexed by triggerGUID
-- same format as an entry in activeEffectsBySpell with two extra fields at the end
-- {..., triggerGUID, refcount}
-- Note that in this approach, a trigger can have only one absorb effect up at the same time.
-- This should be a reasonable assumption for quite some time, considering only Anti-Magic Zone
-- and (soon) Power Word: Barrier use this feature anyway
local activeAreaEffects = {};

-- Table of current unit charges
-- A charge is a variant value a custom trigger can put on any unit with a limited lifetime
-- It is organized as a (very simple) queue (FIFO) for use with Divine Aegis e.g. to save the
-- critical heal value and then apply it on the aura gain or with Val'anyr.
-- [GUID] = { [spellId] = { charge1, charge2, ... } }
local activeCharges = {};

-- Table of known spells that cause an absorb effects
-- priority: effects above 2 are neither used in total value nor displayed (e.g. Anti-Magic Shell)
-- [spellId] = {priority, duration, type, createFunc, hitFunc}
-- actually AbsorbsMonitor.Effects, defined in Effects.lua
local AbsorbEffects;

-- Table of spells that cause an area effect
-- [spellId (trigger)] = spellId (absorb effect)
-- actually AbsorbsMonitor.AreaTriggers, defined in Effects.lua
local AreaTriggers;

-- Table of additional callbacks on combat log events for proc-based and other non-generic absorb effects.
-- actually AbsorbsMonitor.CombatTriggers, defined in Effects.lua
local CombatTriggers;

-- Table of all unit stats relevant to absorb effects like attack power and spell power
-- (mastery rating later on)
-- [GUID] = { class, AttackPower, SpellPower, quality }
AbsorbsMonitor.UnitStats = {};

-- Table of all scaling factors to absorb effects like talents, items, set boni, buffs
-- If there is no mechanism in Cataclysm to obtain the correct absorb amount by any effect
-- this entries are meant to be distributed among a group, raid, etc
-- [GUID] = { [scaling_Name] = scaling_Value }
-- A numerical scaling_name above 10 should ONLY be used if it is the spellId of the affected effect,
-- since it is sometimes used as a very quick way to check for a unit's class
-- Scaling factors that are only relevant to the local user like priest talents and do not
-- need to be distributed but are needed for calculation of the public one's are private ones,
-- found at index -1
AbsorbsMonitor.Scaling = {[-1] = {}};

-- Class-specific callbacks
AbsorbsMonitor.OnEnableClass = {};
AbsorbsMonitor.OnScalingDecode = {};

-- Table for spell names
-- This one gets populated automatically on each aura application
AbsorbsMonitor.SpellNames = {
	[0] = (LibStub("AceLocale-3.0"):GetLocale("AbsorbsMonitor"))["Unknown"],
}

-- Enum for effect type
AbsorbsMonitor.EffectType	 = {	
	BUFF = 1,			-- An absorb effect gained by application of a (buff) aura on another unit, only
						-- created for this unit with a limited duration (e.g. Power Word: Shield, ...)
						
	AREA = 2,			-- An absorb effect gained by a summoned unit that radiates an aura to other units
						-- within its vicinity (e.g. Anti-Magic Zone, Power Word: Barrier, ...)
						
	PASSIVE = 3			-- A passive effect (usually by talents) that may kick in at certain points like
						-- low health (e.g. Ardent Defender, Cheat Death, ...)
};

local EffectTypeBUFF = AbsorbsMonitor.EffectType.BUFF;
local EffectTypeAREA = AbsorbsMonitor.EffectType.AREA;
local EffectTypePASSIVE = AbsorbsMonitor.EffectType.PASSIVE;



-- #################################################################################################


local CommStatsCooldown = false;

local function ClearCommStatsCooldown()
	CommStatsCooldown = false;
end

local CommScalingCooldown = false;

local function ClearCommScalingCooldown()
	CommScalingCooldown = false;
end

local function DummyFunc()
end

local function DeepTableCopy(src)
	dest = {};
	
	for k, v in pairs(src) do
		if(type(k) == "table") then
			k = DeepTableCopy(k);
		end
		
		if(type(v) == "table") then
			v = DeepTableCopy(v);
		end
		
		dest[k] = v;
	end
	
	setmetatable(dest, getmetatable(src));
	
	return dest;
end

local function SortEffects(a, b)
	if(a[1] == b[1]) then
		return (a[3] < b[3]);
	else
		return (a[1] < b[1]);
	end
end


local ApplyEffectFunc = {

	[EffectTypeBUFF] = function(self, sourceGUID, sourceName, destGUID, destName, spellId)
		local destEffects = activeEffectsBySpell[destGUID];
		local effectInfo = AbsorbEffects[spellId];
		local effectEntry;
	
		-- To help me find typos
		if(not effectInfo[4]) then
			self:Print("create func nil on spellId "..spellId);
		end
		
		local value, quality, extra = effectInfo[4](sourceGUID, sourceName, destGUID, destName, spellId, destEffects);
	
		if(value == nil) then return; end
	
		-- no entry yet for this unit
		if(not destEffects) then
			effectEntry = {spellId, effectInfo[1], value, value, quality, 0, extra};
			
			destEffects = {[-1] = 0, [-2] = 1.0, [spellId] = effectEntry};
			
			activeEffectsBySpell[destGUID] = destEffects
			activeEffectsByPriority[destGUID] = {effectEntry};
			
			self:Message_EffectApplied(sourceGUID, sourceName, destGUID, destName, spellId, value, quality);
		
		-- not this specific effect yet
		elseif(not destEffects[spellId]) then	
			effectEntry = {spellId, effectInfo[1], value, value, quality, 0, extra};
					
			destEffects[spellId] = effectEntry;
		
			table.insert(activeEffectsByPriority[destGUID], effectEntry);
			table.sort(activeEffectsByPriority[destGUID], SortEffects);
			
			self:Message_EffectApplied(sourceGUID, sourceName, destGUID, destName, spellId, value, quality);
		
		-- effect exists already
		else	
			effectEntry = destEffects[spellId];
			local prevAmount = effectEntry[3];
		
			effectEntry[3] = value;
			effectEntry[4] = value;
			effectEntry[5] = quality;
			effectEntry[7] = extra;
		
			self:Message_EffectUpdated(destGUID, spellId, value, quality);
		
			-- adjust value in case this is a visible absorb
			-- to get the difference
			value = value - prevAmount;
			
			-- Cancel the exting duration timeout timer
			self:CancelTimer(effectEntry[6], true);
		end
		
		if(quality < destEffects[-2]) then
			destEffects[-2] = quality;
		end
		
		if(effectInfo[1] < 2) then
			destEffects[-1] = destEffects[-1] + value;
		
			self:Message_UnitUpdated(destGUID, destEffects[-1], destEffects[-2]);
		end
	
		-- We add a 5s grace period for latency and all kind of stuff
		-- This duration timeout should only be needed if the unit moved out of combat log
		-- reporting range anyway
		effectEntry[6] = self:ScheduleTimer("OnBuffTimeout", effectInfo[2] + 5, {destGUID, spellId});
		
		--self:Print("applied buff effect "..spellId.." (priority: "..effectInfo[1]..") on "..destGUID.." for "..value..", new total: "..activeEffectsBySpell[destGUID][0].." ("..#(activeEffectsByPriority[destGUID])..")");
	end,

	[EffectTypeAREA] = function(self, triggerGUID, triggerName, destGUID, destName, spellId)	
		if(not activeAreaEffects[triggerGUID]) then
			return self:ApplyBuffEffect(triggerGUID, triggerName, destGUID, destName, 0);
		end
		
		local effectEntry = activeAreaEffects[triggerGUID];			
		local destEffects = activeEffectsBySpell[destGUID];
		
		if(not destEffects) then
			-- Quality of 1.1 to enforce message
			destEffects = {[-1] = 0, [-2] = 1.1};
		
			activeEffectsBySpell[destGUID] = destEffects;
			activeEffectsByPriority[destGUID] = {};
			
		-- At the moment it is impossible for such an effect to be refreshed
		-- Since it is created by a summoned unit radiating it, it either
		-- gets removed/reapplied or removed/applied by a different unit.					
		elseif(destEffects[spellId]) then		
			error("Called ApplyAreaEffect on refreshed aura");
			return;
		end
		
		destEffects[spellId] = effectEntry;				
		effectEntry[9] = effectEntry[9] + 1; -- increase refcount
		
		-- While we keep the trigger itself as a normal afflicted unit to keep
		-- track when the area effect breaks, we do not broadcast it nor have to
		-- keep a sorted priority list
		if(triggerGUID ~= destGUID) then
			table.insert(activeEffectsByPriority[destGUID], effectEntry);
			table.sort(activeEffectsByPriority[destGUID], SortEffects);		
		
			-- Note that we CANNOT use nil as an amount, since external addons can rely on this value being non-nil
			-- for sorting. By using -1, a "simple" addon can sort it quickly, yet area effects will always appear
			-- as top priority
			self:Message_EffectApplied(triggerGUID, triggerName, destGUID, destName, spellId, -1, effectEntry[5]);
			
			-- Update quality if needed
			if(effectEntry[5] < destEffects[-2]) then
				destEffects[-2] = effectEntry[5];
			
				self:Message_UnitUpdated(destGUID, destEffects[-1], destEffects[-2]);
			end
		end
		
		--self:Print("applied area effect "..spellId.." (priority: "..effectData[1]..") on "..destGUID.." for "..value..", new total: "..activeEffectsBySpell[destGUID][0].." ("..#(activeEffectsByPriority[destGUID])..")");
	end,

	[EffectTypePASSIVE] = function(self, sourceGUID, sourceName, destGUID, destName, spellId)
	end
};


function AbsorbsMonitor:CreateAreaTrigger(sourceGUID, sourceName, triggerGUID, triggerName, spellId)
	if(activeAreaEffects[triggerGUID]) then
		error("Trying to create new area trigger on existing one, triggerGUID: "..triggerGUID..", existing spellId: "..activeAreaEffects[triggerGUID][1]);
		return;
	end		

	local effectInfo = AbsorbEffects[spellId];
	
	local value, quality, extra = effectInfo[4](sourceGUID, sourceName, destGUID, destName, spellId, nil);
			
	if(value == nil) then return; end
	
	--self:Print("Creating area tigger "..triggerGUID.."/"..triggerName.." from "..sourceName.." with spellId "..spellId.." for "..value.."/"..quality);	
	
	local effectEntry = {spellId, effectInfo[1], value, value, quality, 0, extra, triggerGUID, 0};
	effectEntry[6] = self:ScheduleTimer("OnAreaTimeout", effectInfo[2] + 5, effectEntry);
	
	activeAreaEffects[triggerGUID] = effectEntry;
	
	self:Message_AreaCreated(sourceGUID, sourceName, triggerGUID, spellId, value, quality);
end

function AbsorbsMonitor:ApplyBuffEffect(...)
	ApplyEffectFunc[EffectTypeBUFF](self, ...);
end

function AbsorbsMonitor:ApplyAreaEffect(...)
	ApplyEffectFunc[EffectTypeAREA](self, ...);
end

function AbsorbsMonitor:ApplyPassiveEffect(...)
	ApplyEffectFunc[EffectTypePASSIVE](self, ...);
end

-- Only meant for ACTIVE (BUFF, AREA) effects
-- Note that this method should NOT be called on non-existing effects or units
-- There are no exist checks within it.
function AbsorbsMonitor:RemoveActiveEffect(guid, spellId)
	local guidEffects = activeEffectsBySpell[guid];
	local effectInfo = AbsorbEffects[spellId];
	
	--self:Print("removing absorb effect "..spellId.." on "..guid.." with "..guidEffects[spellId][3].." left, total: "..guidEffects[-1].." ("..#(activeEffectsByPriority[guid])..")");
			
	if(effectInfo[3] == EffectTypeBUFF) then
		if(effectInfo[1] < 2) then
			guidEffects[-1] = guidEffects[-1] - guidEffects[spellId][3];		
			
			self:Message_UnitUpdated(guid, guidEffects[-1], guidEffects[-2]);
		end
	
		self:CancelTimer(guidEffects[spellId][6], true);
		
		self:Message_EffectRemoved(guid, spellId);	
	
	-- only AREA possible for now
	else
		local effectEntry = guidEffects[spellId];
		
		effectEntry[9] = effectEntry[9] - 1;
		
		if(guid ~= effectEntry[8]) then
			self:Message_EffectRemoved(guid, spellId);
		end
		
		if(effectEntry[9] == 0) then
			if(effectEntry[6]) then
				self:CancelTimer(effectEntry[6], true);
			end
			
			self:Message_AreaCleared(effectEntry[8]);
		end
	end		
	
	if(#(activeEffectsByPriority[guid]) == 1) then		
		activeEffectsBySpell[guid] = nil;
		activeEffectsByPriority[guid] = nil;
				
		self:Message_UnitCleared(guid);
	else
		guidEffects[spellId] = nil;
		
		for k, v in pairs(activeEffectsByPriority[guid]) do
			if(v[1] == spellId) then
				table.remove(activeEffectsByPriority[guid], k);
				
				break;
			end
		end
	end
end

-- Local variable to speed up access
local lowValueTolerance;

function AbsorbsMonitor:SetLowValueTolerance(value)
	lowValueTolerance = value;
end

function AbsorbsMonitor:HitUnit(guid, absorbedTotal, overkill, spellSchool)
	if(not activeEffectsBySpell[guid]) then return; end
	
	local guidEffects = activeEffectsBySpell[guid];
	
	local absorbedRemaining = absorbedTotal;
	
	local absorbed = 0;
	local keepEffect, visibleAbsorb = false, false;
	local i = 1;
	local effectEntry, effectInfo;
	
	-- This loop lasts as long as there is still an absorb value that
	-- no effect could account for, but it will break once the list of
	-- available effects got used completely.
	while(absorbedRemaining > 0) do
		effectEntry = activeEffectsByPriority[guid][i];
		
		-- Sometimes there can be holes in this list since we don't re-sort
		-- after removing an effect
		if(effectEntry == nil) then break; end
		
		effectInfo = AbsorbEffects[effectEntry[1]];
		
		if(effectEntry[3] > 0) then
			-- Hit the abosrb effect
			absorbed, keepEffect = effectInfo[5](guid, effectEntry, absorbedRemaining, overkill, spellSchool);
			
			if(absorbed > 0) then
				-- Reduce the value of this effect and the remaining absorb value
				-- to be accounted for
				effectEntry[3] = effectEntry[3] - absorbed;
				absorbedRemaining = absorbedRemaining - absorbed;

				if(effectInfo[3] == EffectTypeBUFF) then
					-- If it should be visible (priority < 2), correct the total value
					if(effectEntry[2] < 2) then
						guidEffects[-1] = guidEffects[-1] - absorbed;
				
						-- Shows us that at least one visible effect got hit
						visibleAbsorb = true;
					end
					
					self:Message_EffectUpdated(guid, effectEntry[1], effectEntry[3]);
					
					--self:Print("Hit effect #"..i..", absorbed "..absorbed.." (keep: "..tostring(keepEffect).."), remaining: "..absorbedRemaining.." - total: "..guidEffects[-1]);
					
				-- only AREA left
				else
					self:Message_AreaUpdated(effectEntry[8], effectEntry[3]);
				
					--self:Print("Hit area effect #"..i..", absorbed "..absorbed.." (keep: "..tostring(keepEffect).."), remaining: "..absorbedRemaining.." - unit: "..effectEntry[3].."/"..activeEffectsBySpell[guid][0]);
				end						
			end
			
			-- If the hit-function told us to remove the effect, do so
			-- Note that only AbsorbsMonitor:RemoveActiveEffect is allowed to
			-- remove it from the list, we just set the value to zero, so it gets
			-- ignored on any hit. This is do not come into any desync issues with
			-- the events, and to keep the proper clean-up code in one place
			if(not keepEffect) then				
				effectEntry[3] = 0;
			end
		end
			
		i = i + 1;
	end	
	
	-- there are two possibilities when things are going wrong
	-- 
	--	a)	we guessed an absorb value too high, in that case it will
	--		automatically be corrected when it breaks
	--
	--	b)	we guessed an absorb value too low, so we end up with an
	--		amount to absorb when all effects seem to be gone
	--		(absorbedRemaining > 0)
	--		Note that we cannot rely on SPELL_AURA_REMOVED to check this,
	--		since it may happen completely out of order, but it will
	--		clear this unit soon or did so already.
	--		we reduce the quality to zero, since any absorb now happening
	--		cannot be accounted for.
	--		Since we may have rounding errors from scanning the spellbook
	--		and calculating the value thereafter (does Blizzard round on
	--		EVERY step?!?), we accept a small threshold
	
	-- TODO: optimize that deep table access
	if(absorbedRemaining > lowValueTolerance) then
		guidEffects[-1] = guidEffects[-1] - absorbedRemaining;
		guidEffects[-2] = 0.0;
		visibleAbsorb = true;
	end
	
	if(visibleAbsorb) then
		self:Message_UnitUpdated(guid, guidEffects[-1], guidEffects[-2]);
	end	
end

-- This is uses a _very_ simple queue implementation with table.insert and
-- table.remove. It will not scale very well for large values, but we're talking
-- of a maximum of ~3 entries per GUID at any given time. A proper implementation
-- with linked list would probably not be any faster.
function AbsorbsMonitor:PushCharge(guid, spellId, amount, lifetime)
	local guidCharges = activeCharges[guid];

	if(not guidCharges) then
		activeCharges[guid] = { [spellId] = {lastCombatLogEvent + lifetime, amount} };
		
	elseif(not guidCharges[spellId]) then
		guidCharges[spellId] = {lastCombatLogEvent + lifetime, amount};
	
	else
		table.insert(guidCharges[spellId], lastCombatLogEvent + lifetime);
		table.insert(guidCharges[spellId], amount);
	end
end

function AbsorbsMonitor:PopCharge(guid, spellId)
	local guidCharges = activeCharges[guid];
	
	if(guidCharges) then
		local queue = activeCharges[guid][spellId];
		
		-- For some weird reason, it will fail (true even on empty array)
		-- if checked for queue[1] ?!?
		if(queue and queue[2]) then
			local chargeAmount;
			local chargeExpire;			
		
			-- This loop will not be able to run infinitely
			while(true) do
				-- In this order we might save one table reshuffle
				chargeAmount = table.remove(queue, 2);
				
				if(not chargeAmount) then return 0; end
				
				chargeExpire = table.remove(queue, 1);
			
				if(chargeExpire > lastCombatLogEvent) then
					return chargeAmount;
				end
			end
		end
	end
	
	return 0;
end

function AbsorbsMonitor:AddCombatTrigger(target, event, func)
	local index = target.."_"..event;
	local oldTrigger = rawget(CombatTriggers, index);

	if(not oldTrigger) then
		CombatTriggers[index] = func;
		
	else
		local listIndex = index.."_list";
		local funcList = rawget(CombatTriggers, listIndex);
		
		-- There is already a list of callbacks, so we just add this one
		if(funcList) then
			for k, v in pairs(funcList) do
				if(v == func) then
					return;
				end
			end
		
			table.insert(funcList, func);
			
		-- We used a direct call so far, create a list and set up a handler
		else
			if(oldTrigger == func) then
				return;
			end
		
			funcList = {oldTrigger, func};			
			CombatTriggers[listIndex] = funcList;
			
			local handler = function(...)
				for k, v in pairs(funcList) do
					v(...);
				end
			end
			
			CombatTriggers[index] = handler;
		end		
	end
end

function AbsorbsMonitor:RemoveCombatTrigger(target, event, func)
	local index = target.."_"..event;
	
	local listIndex = index.."_list";
	local funcList = rawget(CombatTriggers, listIndex);
	
	-- We have a list of callbacks, reduce if possible
	if(funcList) then
		if(#funcList == 2) then
			CombatTriggers[index] = (funcList[1] == func) and funcList[2] or funcList[1];
			CombatTriggers[listIndex] = nil;
		else
			-- ATTENTION: We have to keep the table in place
			-- because the handler references this table
			
			local old_funcList = DeepTableCopy(funcList);
			wipe(funcList);
			
			for k, v in pairs(old_funcList) do
				if(v ~= func) then				
					table.insert(funcList, v);
				end
			end
		end
	
	-- It was a direct call anyway
	else
		CombatTriggers[index] = nil;
	end	
end

function AbsorbsMonitor:Unit_Stats(guid, missingQuality)
	local guidStats = self.UnitStats[guid];
	
	if(guidStats) then
		return guidStats[2], guidStats[3], guidStats[4];
	else
		return 0, 0, missingQuality;
	end		
end

function AbsorbsMonitor:Unit_Scaling(guid, defaultScaling, defaultQuality)
	local guidScaling = self.Scaling[guid];

	if(guidScaling) then
		return guidScaling, 1.0;
	else
		return defaultScaling, defaultQuality;
	end
end

-- Optimized method to save one function call on creation, since a lot of spells
-- actually require stats and scaling
function AbsorbsMonitor:Unit_StatsAndScaling(guid, missingQuality, defaultScaling, defaultQuality)
	local guidStats = self.UnitStats[guid];
	local guidScaling = self.Scaling[guid];
	
	if(guidStats) then
		if(guidScaling) then
			return guidStats[2], guidStats[3], guidStats[4], guidScaling, 1.0;
		else
			return guidStats[2], guidStats[3], guidStats[4], defaultScaling, defaultQuality;
		end
	else
		if(guidScaling) then
			return 0, 0, missingQuality, guidScaling, 1.0;
		else
			return 0, 0, missingQuality, defaultScaling, defaultQuality;
		end
	end
end

function AbsorbsMonitor:ScheduleScalingBroadcast()
	if(curChatChannel) then
		self:ScheduleUniqueTimer("comm_scaling", "Comm_Scaling", CommScalingCooldown and 30 or 5);
	end
end

-- An extension of AceTimer to schedule a one-shot timer that will not
-- be scheduled twice if scheduled again before it's fired.
-- Note that the callback itself has to make sure the timer gets marked
-- as done
-- I know this could also be done with a custom callback clearing it,
-- but then I have to mimic AceTimer's API concerning the callback
local activeTimers = {};

function AbsorbsMonitor:ScheduleUniqueTimer(id, callback, delay, arg)
	if(not activeTimers[id]) then
		self:ScheduleTimer(callback, delay, arg);
			
		activeTimers[id] = 1;
	end
end

function AbsorbsMonitor:FinishUniqueTimer(id)
	activeTimers[id] = nil;
end

function AbsorbsMonitor:SetActive()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnCombatLogEvent");		
	
	self.db.profile.passive = false;
end

function AbsorbsMonitor:SetPassive()
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	self.db.profile.passive = true;
end

function AbsorbsMonitor:SetVerbose()
	self:RegisterEvent("PARTY_MEMBERS_CHANGED", "OnPartyEvent");
	self:RegisterEvent("RAID_ROSTER_UPDATE", "OnPartyEvent");
	self.db.profile.silent = false;
	
	self:OnPartyEvent();
end

function AbsorbsMonitor:SetSilent()
	self:UnregisterEvent("PARTY_MEMBERS_CHANGED");
	self:UnregisterEvent("RAID_ROSTER_UPDATE");
	self.db.profile.silent = true;
end

function AbsorbsMonitor:OnBuffTimeout(args)
	local guid, spellId = args[1], args[2];

	if(activeEffectsBySpell[guid] and activeEffectsBySpell[guid][spellId]) then
		self:RemoveActiveEffect(guid, spellId);
	end
end

-- This should be called in even less cases than the normal timeout
function AbsorbsMonitor:OnAreaTimeout(areaEntry)
	-- Disable the timer handle entry already
	areaEntry[6] = nil;

	for guid, guidEffects in pairs(activeEffectsBySpell) do
		for spellId, effectEntry in pairs(guidEffects) do
			if(effectEntry == areaEntry) then				
				self:RemoveActiveEffect(guid, spellId);
				
				if(areaEntry[9] == 0) then return; end
			end
		end
	end
	
	-- We're only here if we didn't reduce the refcount
	-- to zero
	error("Positive refcount "..areaEntry[9].." remained for area effect "..areaEntry[1].." by trigger "..areaEntry[8]);
end

-- After profiling this method over several raiding nights, the custom triggers increased
-- the execution time (without its called functions) on average from 0.7 ys to 2.8 ys.
-- Given we have less than 100 combat log events happening per second, it's an acceptable
-- tradeoff given the features we get
function AbsorbsMonitor:OnCombatLogEvent(_, timestamp, type, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	lastCombatLogEvent = timestamp;

	if(type == "SWING_MISSED") then
		local missType, amountMissed = select(1, ...);
		
		if(missType ~= "ABSORB") then return; end	
		
		self:HitUnit(destGUID, amountMissed, 0, PHYSICAL_SPELL_SCHOOL);
		
	elseif(type == "RANGE_MISSED" or type == "SPELL_MISSED" or type == "SPELL_PERIODIC_MISSED") then
		local spellId, spellName, spellSchool, missType, amountMissed = select(1, ...);
			
		if(missType ~= "ABSORB") then return; end
		
		self:HitUnit(destGUID, amountMissed, 0, spellSchool);
		
	elseif(type == "SWING_DAMAGE") then
		local amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
			
		if(not absorbed) then return; end
		
		self:HitUnit(destGUID, absorbed, amount, PHYSICAL_SPELL_SCHOOL);
		
	elseif(type == "RANGE_DAMAGE" or type == "SPELL_DAMAGE" or type == "SPELL_PERIODIC_DAMAGE") then
		local spellId, spellName, spellSchool, amount, overkill, school, resisted, blocked, absorbed = select(1, ...);
		
		if(not absorbed) then return; end
		
		self:HitUnit(destGUID, absorbed, amount, spellSchool);
		
	elseif(type == "SPELL_HEAL") then
		local spellId, spellName, spellSchool, amount, overhealing, absorb, critical = select(1, ...);		
		
		CombatTriggers[sourceGUID.."_OnHeal"](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		
		if(critical) then
			CombatTriggers[sourceGUID.."_OnHealCrit"](sourceGUID, sourceName, destGUID, destName, spellId, amount, overhealing);
		end
	elseif(type == "SPELL_AURA_APPLIED") then
		local spellId, spellName = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			self.SpellNames[spellId] = spellName;
		
			ApplyEffectFunc[AbsorbEffects[spellId][3]](self, sourceGUID, sourceName, destGUID, destName, spellId);
		end
		
		CombatTriggers[spellId.."_OnAuraApplied"](sourceGUID, sourceName, destGUID, destName, spellId);
		
	elseif(type == "SPELL_AURA_REFRESH") then
		local spellId, spellName = select(1, ...);
		
		if(AbsorbEffects[spellId]) then
			self.SpellNames[spellId] = spellName;		
		
			ApplyEffectFunc[AbsorbEffects[spellId][3]](self, sourceGUID, sourceName, destGUID, destName, spellId);
		end	
		
		CombatTriggers[spellId.."_OnAuraApplied"](sourceGUID, sourceName, destGUID, destName, spellId);
				
	elseif(type == "SPELL_AURA_REMOVED") then
		local spellId = select(1, ...);			
		
		if(AbsorbEffects[spellId]) then
			if(activeEffectsBySpell[destGUID] and activeEffectsBySpell[destGUID][spellId]) then
				self:RemoveActiveEffect(destGUID, spellId);
			end
		end
		
		CombatTriggers[spellId.."_OnAuraRemoved"](sourceGUID, sourceName, destGUID, destName, spellId);
		
	elseif(type == "SPELL_SUMMON") then
		local spellId = select(1, ...);
		
		if(AreaTriggers[spellId]) then
			self:CreateAreaTrigger(sourceGUID, sourceName, destGUID, destName, AreaTriggers[spellId]);
		end
		
		--CombatTriggers(spellId.."_OnSummon"](sourceGUID, sourceName, destGUID, destName, spellId);
		
	end
end

function AbsorbsMonitor:OnPartyEvent(...)
	-- Note that the order here is VERY important
	if(UnitInBattleground("player")) then	
		curChatChannel = "BATTLEGROUND";
	elseif(UnitInRaid("player")) then
		curChatChannel = "RAID";
	elseif(GetNumPartyMembers() > 0) then
		curChatChannel = "PARTY";
	else
		curChatChannel = nil;
	end	
end

-- Come on, this whole stuff is bullshit anyway
-- and will hopefully be removed in Cataclysm
function AbsorbsMonitor:OnZoneChangedNewArea()
	local profile = self.db.profile;
	
	if(not profile.zoneModifier_auto) then return; end

	if(UnitInBattleground("player")) then
		profile.zoneModifier = 1.17;
	elseif(IsActiveBattlefieldArena()) then
		profile.zoneModifier = 0.9;
	elseif(GetRealZoneText() == "Wintergrasp") then
		-- 1.3 * 0.9 = fail Blizzard
		profile.zoneModifier = 1.17;
	elseif(GetRealZoneText() == "Icecrown Citadel") then
		profile.zoneModifier = 1.3;
	else
		profile.zoneModifier = 1.3;
	end
end

function AbsorbsMonitor:OnUnitStatsReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, class, ap, sp = self:Deserialize(text);
	
	if(not(success and guid and class and ap and sp)) then return; end
	
	if(guid == playerGUID) then return; end
	
	if(not self.UnitStats[guid]) then
		self.UnitStats[guid] = {class, ap, sp, 1.0};
	else
		self.UnitStats[guid][2] = ap;
		self.UnitStats[guid][3] = sp;
		self.UnitStats[guid][4] = 1.0;
	end
end

function AbsorbsMonitor:OnScalingEncode()
	return self.Scaling[playerGUID];
end

AbsorbsMonitor.OnScalingDecode.DEFAULT = function(self, guid, guidScaling)
	self.Scaling[guid] = guidScaling;
end

function AbsorbsMonitor:OnScalingReceived(prefix, text, distribution, target)
	if(not text) then return; end
	
	local success, guid, class, inScaling = self:Deserialize(text);		
	
	if(not(success and guid and class and inScaling)) then return; end
	
	--if(guid == playerGUID) then return; end
	
	(self.OnScalingDecode[class])(self, guid, inScaling);
end

function AbsorbsMonitor:OnInitialize()
	self:InitOptions();
	
	-- To actually see syntax errors in Effects.lua
	if(not self.CombatTriggers) then return; end
	
	setmetatable(self.CombatTriggers, {
		__index = function()
			return DummyFunc;
		end
	});
	
	setmetatable(self.OnScalingDecode, {
		__index = function(table, class)
			return table.DEFAULT;
		end
	}); 
	
	AbsorbEffects = self.Effects;
	CombatTriggers = self.CombatTriggers;
	AreaTriggers = self.AreaTriggers;
end

function AbsorbsMonitor:OnEnable()
	_, playerClass = UnitClass("player");
	 playerGUID = UnitGUID("player");
	
	self.UnitStats[playerGUID] = {playerClass, 0, 0, 1.0};	
	self.Scaling[playerGUID] = {};	
	
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnZoneChangedNewArea");
	self:OnZoneChangedNewArea();

	if(playerClass == "DEATHKNIGHT") then
		self:RegisterEvent("UNIT_ATTACK_POWER", "OnUnitStatsChanged");
		
	elseif(playerClass == "DRUID") then
		self:RegisterEvent("UNIT_ATTACK_POWER", "OnUnitStatsChanged");
		
	elseif(playerClass == "MAGE") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnUnitStatsChanged");
		
	elseif(playerClass == "PALADIN") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnUnitStatsChanged");
		
	elseif(playerClass == "PRIEST") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnUnitStatsChanged");
		
	elseif(playerClass == "WARLOCK") then
		self:RegisterEvent("PLAYER_DAMAGE_DONE_MODS", "OnUnitStatsChanged");
	end
	
	self:OnUnitStatsChanged();
	
	-- On fresh login, talent information are not yet available
	if(not GetTalentInfo(1, 1)) then
		self:RegisterEvent("PLAYER_ALIVE",
			function()
				self:OnFinalEnable();
				self:UnregisterEvent("PLAYER_ALIVE");
			end
		);
	else
		self:OnFinalEnable();
	end
end

function AbsorbsMonitor:OnFinalEnable()	
	self:EnableEffects();
	
	-- This has to happen before class init, else we get no initial scaling broadcast
	if(not self.db.profile.silent) then
		self:SetVerbose();
	end	
	
	if(self.OnEnableClass[playerClass]) then
		self.OnEnableClass[playerClass](self);
	end
	
	if(self.OnLevelUp) then
		self:RegisterEvent("PLAYER_LEVEL_UP", "OnLevelUp");
	end
	
	if(self.OnTalentUpdate) then
		self:RegisterEvent("PLAYER_TALENT_UPDATE", "OnTalentUpdate");
	end
	
	if(self.OnGlyphUpdated) then
		self:RegisterEvent("GLYPH_UPDATED", "OnGlyphUpdated");
	end
	
	if(self.OnEquipmentChanged) then
		self:RegisterEvent("PLAYER_EQUIPMENT_CHANGED", "OnEquipmentChanged");
	end
	
	self:RegisterComm("Absorbs_UnitStats", "OnUnitStatsReceived");
	self:RegisterComm("Absorbs_Scaling", "OnScalingReceived");	
		
	if(not self.db.profile.passive) then
		self:SetActive();
	end
	
	self:SetLowValueTolerance(self.db.profile.lowValueTolerance);	
	
	--self:Print("Enabled");	
end

function AbsorbsMonitor:OnDisable()
	--self:Print("Disabled");
end

function AbsorbsMonitor:OnUnitStatsChanged(event, unit)
	local baseAP, plusAP, minusAP = UnitAttackPower("player");

	self.UnitStats[playerGUID][2] = baseAP + plusAP - minusAP;
	-- TODO: What about spell power ~= healing spell power?
	self.UnitStats[playerGUID][3] = GetSpellBonusHealing();

	if(curChatChannel) then
		self:ScheduleUniqueTimer("comm_stats", "Comm_UnitStats", CommStatsCooldown and 15 or 5);
	end
end

local lastAP, lastSP = 0, 0;

function AbsorbsMonitor:Comm_UnitStats()
	self:FinishUniqueTimer("comm_stats");

	if(curChatChannel) then
		local curAP, curSP = self.UnitStats[playerGUID][2],  self.UnitStats[playerGUID][3];
		
		if( (curAP ~= lastAP) or (curSP ~= lastSP) ) then
			self:SendCommMessage("Absorbs_UnitStats", self:Serialize(playerGUID, playerClass, curAP, curSP), curChatChannel);
			
			lastAP, lastSP = curAP, curSP;
			
			CommStatsCooldown = true;
			self:ScheduleTimer(ClearCommStatsCooldown, 15);
		end				
	end
end

function AbsorbsMonitor:Comm_Scaling()
	self:FinishUniqueTimer("comm_scaling");

	if(curChatChannel) then
		--self:Print("Sending: ", playerGUID, playerClass, self:OnScalingEncode(self));
		--self:Print("Encoded: "..self:Serialize(playerGUID, playerClass, self:OnScalingEncode(self)));
	
		self:SendMessage("Absorbs_Scaling", self:Serialize(playerGUID, playerClass, self:OnScalingEncode(self)), curChatChannel);
		
		CommScalingCooldown = true;
		self:ScheduleTimer(ClearCommScalingCooldown, 30);
	end
end

-- The effect-individual messages get sent on visible and non-visible effects
-- sourceGUID, sourceName, destGUID, destName, spellId, value, quality
function AbsorbsMonitor:Message_EffectApplied(...)
	self:SendMessage("Absorbs_EffectApplied", ...);
end

-- guid, spellId, value
function AbsorbsMonitor:Message_EffectUpdated(...)
	self:SendMessage("Absorbs_EffectUpdated", ...);
end

-- guid, spellId
function AbsorbsMonitor:Message_EffectRemoved(...)
	self:SendMessage("Absorbs_EffectRemoved", ...);
end

-- Whenever the unit that radiates an AREA effect is created (visible/non-visible)
-- Note that the actual effect on the unit that absorbs damage casues an
-- EffectApplied/EffectRemoved message, but not EffectUpdated (instead AreaUpdated)

-- The rationale behind this is performance, since we cannot update every unit afflicted
-- by the area effect on every hit. Therefore, we have the shared entry in the activeEffects
-- table of each unit, and will handle it the same way when exporting - separately from each
-- others.
-- sourceGUID, sourceName, triggerGUID, spellId, value, quality
function AbsorbsMonitor:Message_AreaCreated(...)
	self:SendMessage("Absorbs_AreaCreated", ...);
end

-- triggerGUID, value
function AbsorbsMonitor:Message_AreaUpdated(...)
	self:SendMessage("Absorbs_AreaUpdated", ...);
end

function AbsorbsMonitor:Message_AreaCleared(triggerGUID)
	self:SendMessage("Absorbs_AreaCleared", triggerGUID);
end

-- Only for VISIBLE changes on the total amount
-- guid, value, quality
function AbsorbsMonitor:Message_UnitUpdated(...)
	self:SendMessage("Absorbs_UnitUpdated", ...);
end

-- Everytime a unit gets cleared from all absorb effects (quality reset)
-- including non-visible effects
function AbsorbsMonitor:Message_UnitCleared(guid)
	self:SendMessage("Absorbs_UnitCleared", guid);
end

function AbsorbsMonitor:PrintProfiling(criticalOnly)
	if(GetCVar("scriptProfile") ~= "1") then
		self:Print("CPU profiling disabled");
		return;
	end

	UpdateAddOnCPUUsage();
	UpdateAddOnMemoryUsage();
	
	self:Print("Mem: "..GetAddOnMemoryUsage("AbsorbsMonitor").." kB");
	
	local funcTable;
	
	if(criticalOnly) then
		funcTable = {
			["ApplyBuffEffect"] = ApplyEffectFunc[EffectTypeBUFF],
			["HitUnit"] = self.HitUnit,
			["RemoveActiveEffect"] = self.RemoveActiveEffect,
			["OnCombatLogEvent"] = self.OnCombatLogEvent,
		};
	else
		funcTable = self;
	end	
		
	local v_type;
	local time_self, time_combined, count;
	
	for k, v in pairs(funcTable) do
		v_type = type(v);
		
		if(v_type == "function") then
			time_self, count = GetFunctionCPUUsage(v, false);
			time_combined = GetFunctionCPUUsage(v, true);
			
			self:Print(k.." (#"..count.."): "..string.format("%.5f", time_self).." / "..string.format("%.5f", time_combined));
		end
	end
end
